/*
   This is a quick fix of e2compress as distributed 1997-04-19.
   A more sophisticated version is still in development.

   1998-02-15: We'll see.  e2compress won't work in e2compr-0.4.x yet.

   */


/* Work in progress:

   Make consider_uncompressed look at cluster bitmap.


   Make decompression work for stdin and non-ext2 filesystems.  make
   iattrs read ioctl not cause abort on failure, but consider it a
   test for whether or not the file is on an ext2 filesystem.  We
   should perhaps read fs/ext2/ioctl.c just to check that this is so
   (e.g. would it fail if we didn't have permission?  But how come we
   have the file open?).  I think we should get rid of the
   .iattrs_valid field of status, and replace it with an `enum
   {ext2_files, non_ext2_files, stdio} .io_type'.

   Look at the compression routines.


   Probably other things.

   */



/* todo: It would be nice to allow -f to allow compressing regardless
   of support for e2compr ioctls, if only so that we can test this
   program and the associated compression routines.

 * TODO: Behave as chattr for directories.  (Not very important; find|xargs will
   do much the same job.)

 * TODO: Clean termination (remove the tmp file!)  (I think this is done now.)

 * TODO: Allow decompressing stdin.
 *
 * TODO: We currently assume that if getflags fails then file is not
 *	 on an ext2 fs.  Check the kernel ioctl code to see if this is
 *	 true.  If so, add it to list of assumptions.
 *
 * todo: Split `--force' into multiple options.
 *
 * todo: Warn if compress compressed file.  (Skip unless -f.)
 *       However, it is useful to run e2compress to change the
 *       compression algorithm and/or cluster size -- particularly if
 *       the current compression algorithm is `none'.
 *
 */

/*
    e2compress.c	Simple program to compress files in a format
			transparently readable by the kernel.  This
  			program turns on the EXT2_NOCOMPR_FL flag in
  			order to have direct access to raw data, and
  			turns it off again when everything is done.
  
  
    The idea and portions of the current code are copyright (C) 1995,
    Antoine Dumesnil de Maricourt.
  
    The remainder is copyright (C) 1996-7, Peter Moulder.  See file
    `CREDITS' for separation, but it is probably unimportant given
    that both Peter and Antoine have placed their work under the GNU
    General Public License.
  

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


    Peter can be contacted at <reiter@netspace.net.au> or
    1/36 Auburn Grove, Hawthorn East 3123, Australia.

    Antoine is no longer the maintainer of the e2compr project so
    should not normally be contacted for maintenance issues (bugs,
    feature requests, etc.), but his e-mail address is
    <dumesnil@etca.fr>.

 */


/* ---=== Assumptions ===---

   This is an incomplete list of assumptions made by this code.  I
   believe that at the time of writing, these hold.  If, as you read
   this, you know that one or more of these does not hold then the
   program most probably needs changes.

   1. The set of valid ext2 filesystem block sizes is {1024, 2048, 4096}.

   2. For compressed clusters with holes, the blocks occupied by the
   compressed data are the first n of the blocks that would be
   occupied by the uncompressed data, where n = round_up_div (length
   of compressed data, filesystem blocksize).

   3. Except for the last cluster in the file, head->len
   (i.e. uncompressed length) equals ,the number of bits raised in
   head->bitmap` multiplied by the filesystem blocksize.
   
   4. Cluster sizes (in bytes) not in {4096,8192,16384,32768} are invalid.

   5. Cluster sizes (in units of filesystem blocks) not in {4,8,16,32}
   are invalid.

   6. E2compression only applies to files on ext2 filesystems.

   7. The EXT2_IOC_SETFLAGS and EXT2_IOC_GETFLAGS ioctls succeed if
   and only if applied to a file on an ext2 filesystem.

  */

/* Define this if you want to enable support for benchmarking.  */
/* Note: at present, the benchmarking code is unusable.  (TODO.) */
#undef BENCH  /* I'll reenable this once benchmarking stuff is worked out. */

#include <linux/autoconf.h>
#ifndef CONFIG_EXT2_COMPRESS
/* A few things need to be defined for <linux/ext2_fs.h> to give
   us what we want. */
# define CONFIG_EXT2_COMPRESS
# define CONFIG_EXT2_DEFAULT_COMPR_METHOD_GZIP9 /* Change this if you want. */
# define CONFIG_EXT2_DEFAULT_CLUSTER_BITS_5     /* Change this if you want. */
# define CONFIG_EXT2_USE_GZIP /* prevents ext2_fs.h from giving #error. */
#endif

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/types.h>
#include <utime.h>
#include <sys/ioctl.h>
#include <linux/fs.h>	/* FIGETBSZ */
#include <sys/stat.h>
#include <sys/mman.h>	/* mmap */
#include <values.h>	/* BITSPERBYTE */
#include <signal.h>
#ifdef BENCH
# include <sys/time.h>
# include <sys/resource.h>
#endif
#include <assert.h>
#include <linux/ext2_fs.h>
#include <bytesex.h>
#include "nio.h"
#include <asm/page.h>  /* PAGE_SIZE */
#include "compat.h"

extern char const *strsignal(int);
typedef unsigned int u32;
typedef unsigned short u16;

typedef enum { false, true } bool;


/* --- Configuration --- */

/* Default compression algorithm and cluster size. */
#define DFL_COMPR_METH EXT2_DEFAULT_COMPR_METHOD 
#define DFL_CLU_NBLOCKS EXT2_DEFAULT_CLUSTER_SIZE
/* Note: EXT2_DEFAULT_COMPR_METHOD and EXT2_DEFAULT_CLUSTER_SIZE are
   defined according to the most recent run of `make config' (or
   similar).  Thus their values will vary from one compile environment
   to another.  I think this is a feature, but you may disagree.

   TODO: A better way would be to ask the kernel for the compression
   method and cluster size once the temp file has been creat'ed and
   had the COMPR attribute set. */


static char const version_string[] = "e2compress 1998-06-20"; /* fixme */

static const char *prog_name;
void *heap_base = NULL;

#ifdef BENCH
static unsigned long cmsec = 0;
static unsigned long umsec = 0;
static unsigned long cbyte = 0;
static unsigned long ubyte = 0;
static unsigned long cblk = 0;
static unsigned long ublk = 0;
#endif

/* Compression algorithm working area size. */
#define HEAP_SIZE	(256 * 1024)

/*
 *	Defined in ext2_fs.h, but under the protection of __KERNEL__
 */

extern size_t ext2_iLZV1   (void*);
extern size_t ext2_iLZV2   (void*);
extern size_t ext2_iLZRW3A (void*);
extern size_t ext2_iNONE   (void*);
extern size_t ext2_iGZIP   (void*);
extern size_t ext2_iBZIP2  (void*);

extern size_t ext2_wLZV1   (__u8*, __u8*, size_t, size_t, int);
extern size_t ext2_wLZV2   (__u8*, __u8*, size_t, size_t, int);
extern size_t ext2_wLZRW3A (__u8*, __u8*, size_t, size_t, int);
extern size_t ext2_wNONE   (__u8*, __u8*, size_t, size_t, int);
extern size_t ext2_wGZIP   (__u8*, __u8*, size_t, size_t, int);
extern size_t ext2_wBZIP2  (__u8*, __u8*, size_t, size_t, int);

extern size_t ext2_rLZV1   (const __u8*, __u8*, size_t, size_t, int);
extern size_t ext2_rLZV2   (const __u8*, __u8*, size_t, size_t, int);
extern size_t ext2_rLZRW3A (const __u8*, __u8*, size_t, size_t, int);
extern size_t ext2_rNONE   (const __u8*, __u8*, size_t, size_t, int);
extern size_t ext2_rGZIP   (const __u8*, __u8*, size_t, size_t, int);
extern size_t ext2_rBZIP2  (const __u8*, __u8*, size_t, size_t, int);

extern unsigned long ext2_adler32 (unsigned long adler, const unsigned char *buf, unsigned len);


/*
 *	Algorithm table.
 */

struct ext2_algorithm { 
  const char    *name; 
  size_t (*init)	(void*); 
  size_t (*compress)	(__u8*, __u8*, size_t, size_t, int); 
  size_t (*uncompress)	(const __u8*, __u8*, size_t, size_t, int);
  int	   xarg;
};

static struct ext2_algorithm algo[32] = {

  { "lzv1"	, ext2_iLZV1	, ext2_wLZV1	, ext2_rLZV1	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { "none"	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { "bzip2"	, ext2_iBZIP2	, ext2_wBZIP2	, ext2_rBZIP2	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},

  { "lzrw3a"	, ext2_iLZRW3A	, ext2_wLZRW3A	, ext2_rLZRW3A	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},

  { "gzip1"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 1	},
  { "gzip2"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 2	},
  { "gzip3"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 3	},
  { "gzip4"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 4	},
  { "gzip5"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 5	},
  { "gzip6"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 6	},
  { "gzip7"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 7	},
  { "gzip8"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 8	},

  { "gzip9"	, ext2_iGZIP	, ext2_wGZIP	, ext2_rGZIP	, 9	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
  { NULL	, ext2_iNONE	, ext2_wNONE	, ext2_rNONE	, 0	},
};
static int highest_known_method_id = EXT2_GZIP9_ID;

/*
 *	Algorithm name to ID conversion (should be shared with chattr)
 */

static struct { const char *name; int id; } name2id[] = {
  { "lzv"	, EXT2_LZV1_ID	 	},
  { "lzv1"	, EXT2_LZV1_ID	 	},
  { "none"	, EXT2_NONE_ID  	},
  { "bzip2"	, EXT2_BZIP2_ID	 	},
  { "lzrw3a"	, EXT2_LZRW3A_ID	},
  { "gzip"	, EXT2_GZIP6_ID	 	},
  { "gzip1"	, EXT2_GZIP1_ID	 	},
  { "gzip2"	, EXT2_GZIP2_ID	 	},
  { "gzip3"	, EXT2_GZIP3_ID	 	},
  { "gzip4"	, EXT2_GZIP4_ID	 	},
  { "gzip5"	, EXT2_GZIP5_ID	 	},
  { "gzip6"	, EXT2_GZIP6_ID	 	},
  { "gzip7"	, EXT2_GZIP7_ID	 	},
  { "gzip8"	, EXT2_GZIP8_ID	 	},
  { "gzip9"	, EXT2_GZIP9_ID	 	},
  { NULL	, 0 }
};

/*
 *	Help message
 */

/* Todo: Update this usage text to show long options.  Also modify -B interface. */
static void volatile 
usage (int rc)
{
  fprintf (stderr, "\
Usage: %s [-b size] [-m method] [-F size] [-f] [-t] [-v] files ...
  %s -d [-F size] [-f] [-v] files ...
  %s -h | --help | --version
Options:
  -F, --force-blocksize <size>
	By default we use an ioctl to determine the blocksize of a file.
	Use this flag to use a different size (specified in bytes).
	It's hard to think of a circumstance where you'd want to use
	this flag.
	A size of 0 means go back to using the ioctl.
  -S, --suffix <suffix>
	Set backup suffix to <suffix>.  (Default=`~z~')
"
#ifdef BENCH
	   "\
  -B --benchmark
	Provide timing information for benchmarking.
  -N --loops <count>
	Set the number of times to loop with -B (default=10).
"
#endif
	   "\
  -f, --force 
	Force: 
	 -  write over existing backup file, 
	 -  continue even if more than one link (the other
	    links will remain unchanged),
	 -  attempt to uncompress even if original isn't flagged as
	    being compressed,
	 -  disables verification of compression
  -b, --cluster-blocks <size>
	Set cluster size (specified as a number of filesystem blocks) for
	compression.
  -d, --decompress, --uncompress
	Decompress files.
  -m, --method, --algorithm <method>
	Set compression algorithm for compression
  --believe-bitmap
	When decompressing, trust the cluster bitmap even if a cluster
	otherwise appears to be a valid compressed cluster.
  -h, --help
	Show this usage message and exit immediately.
  -v, --verbose
	Display cluster statistics.
  --version
	Display the version string and exit immediately.
",
	   prog_name, prog_name, prog_name);
  exit (rc);
}


#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))


#ifndef cpu_to_le16
# if __BYTE_ORDER == 1234
#  define cpu_to_le16(x) (x)
#  define cpu_to_le32(x) (x)
#  define le16_to_cpu(x) (x)
#  define le32_to_cpu(x) (x)
# elif __BYTE_ORDER == 4321
#  define cpu_to_le16(x) swab16(x)
#  define cpu_to_le32(x) swab32(x)
#  define le16_to_cpu(x) swab16(x)
#  define le32_to_cpu(x) swab32(x)
static inline u16
swab16 (u16 x)
{
  return ((x & 0xff) << 8) | (x >> 8);
}

static inline u32
swab32 (u32 x)
{
  return ((swab16 (x & 0xffff) << 16)
	  | swab16 (x >> 16));
}
# else
#  error "Unrecognised byte ordering."
# endif
#endif


static inline bool
more_than_one_bit_set (unsigned long x)
{
  return (x & (x - 1));
}

#define is_power_of_two(x) (!more_than_one_bit_set(x))

static bool
valid_blocksizes (unsigned fsblk_min_nchars, unsigned fsblk_max_nchars)
{
  return (is_power_of_two(fsblk_min_nchars)
	  && is_power_of_two(fsblk_max_nchars)
	  && (fsblk_min_nchars >= EXT2_MIN_BLOCK_SIZE)
	  && (fsblk_max_nchars <= EXT2_MAX_BLOCK_SIZE)
	  && (fsblk_min_nchars <= fsblk_max_nchars));
}

static bool
valid_blocksize (unsigned fsblk_nchars)
{
  return (is_power_of_two(fsblk_nchars)
	  && (fsblk_nchars >= EXT2_MIN_BLOCK_SIZE)
	  && (fsblk_nchars <= EXT2_MAX_BLOCK_SIZE));
}

/* static bool valid_clu_nchars (unsigned, unsigned) __attribute__ ((unused)); */

static bool
valid_clu_nchars (unsigned x, unsigned fsblk_min_nchars)
{
  return (is_power_of_two(x)
	  && (x <= 32*1024)
	  && (x >= 4 * fsblk_min_nchars));
}

static char const *
xbasename (char const *full)
{
  char const *p = strrchr (full, '/');
  if (p)
    return p + 1;
  else
    return full;
}


#define strcmpi(x,y) strcasecmp ((x), (y))

#define streq(x, y) (strcmp ((x), (y)) == 0)


static inline unsigned long round_up_div (unsigned long, unsigned) __attribute__ ((const));
static inline unsigned long round_up (unsigned long, unsigned) __attribute__ ((const));

static inline unsigned long
round_up (unsigned long x, unsigned y)
{
  assert (y > 0);
  assert (is_power_of_two (y));
  return ((x - 1) | (y - 1)) + 1;
}

static inline unsigned long
round_up_div (unsigned long num, unsigned den)
{
  if (num)
    return ((num - 1) / den) + 1;
  else
    return 0;
}

static inline bool
round_up_ge(unsigned long a, unsigned long b, unsigned n)
{
  unsigned mask = n - 1;

  assert (is_power_of_two (n));
  assert (n != 0);
  return ((a >= b)  /* This subtest is to save a bit of arithmetic. */
  	  || ((a != 0)  /* Necessary because of unsigned arithmetic. */
	      && ((a - 1) | mask) >= ((b - 1) | mask)));
}

static unsigned
count_bits (unsigned long x)
{
  register unsigned n = 0;

  while ((x & 0x3ff) == 0x3ff)
    {
      n += 10;
      x >>= 10;
    }

  while ((x & 0xf) == 0xf)
    {
      n += 4;
      x >>= 4;
    }
  
  while (x)
    {
      n++;
      x &= (x - 1);
    }
  return n;
}


static bool
is_zero (void const * const o, size_t len)
     /* Returns true iff at least len bytes beginning at o are NULs.
	(Returns true if len == 0.) */
{
  if (len < sizeof (unsigned long))
    {
      register char const *p = (char const *) o;

      while (len--)
	if (*p++)
	  return false;
      return true;
    }
  else
    {
      return !(*((unsigned long const *) o)
	       || memcmp ((const void *) o,
			  (const void *) (((unsigned long const *) o) + 1),
			  len - sizeof (unsigned long)));
    }
}


#if 0
/* The following is more portable than the above in that it doesn't require
   casting a char* to a unsigned*. */
static bool
is_zero (char const *p, size_t len)
{
  assert (len > 0);
  return !(*p++
	   || (--len
	       && (*p++
		   || (--len
		       && (*p++
			   || (--len
			       && (*p++
				   || memcmp (o, p, --len))))))));
}
#endif

enum action {ACT_decompress, ACT_compress, ACT_check};

typedef struct file_status {
  /* Command-line-specified. */
  char const   *ipath;
  char const   *opath;
  unsigned	fs_blk_size;
  unsigned	clu_nblocks;
  int		compr_meth;
  enum action	action;
  bool		verbose;
  bool		stress_test;
  bool		do_continue;
  bool		believe_bitmap;
  bool		to_stdout;
#ifdef BENCH
  bool		do_bench;
  int		nloop;
#endif

  /* Other. */
  off_t		ilen;
  umode_t	imode;
  bool		restore_iattrs;
  /* The file_error() routine checks this to see whether it should
     restore the ext2 attributes of .ifd to those in .iattrs.
     Initially this flag is false.  Just before changing the
     attributes on the input file, we fill .iattrs, and set
     restore_iattrs to true. */
  struct {
    int iattrs:1;
    int clu_nblocks:1;
  } 		valid;  /* 0 bits indicate an unitialised field. */
  unsigned long	iattrs;
  /* .iattrs holds the original ext2 attributes from the file
     referenced by .ifd. */
#if 0
  enum {	/* Note that the following are ordered, so we can use `<' etc. */
    UNSET,  		/* We haven't found out yet. */
    stdio,  		/* Copying from stdin to stdout. */
    non_ext2_files,	/* Timestamps, owner, group, permissions may
			   or may not work.  Ext2 attributes and
			   e2compression are not expected to work. */
    ext2_uncmp_files,	/* Ext2 filesystem but not e2compressed.
			   (User gave `-f' flag.) */
    ext2_e2cmp_files	/* Everything should work (so long as the
			   kernel has e2compr support compiled in). */
  }		io_type;
#endif
  uid_t 	i_uid;
  gid_t 	i_gid;
  int		ifd;
  int		ofd;
  struct utimbuf timestamps;
  unsigned	clu_nchars;	/* Equals (fs_blk_size * clu_nblocks). */
  bool		file_e2compr_capable;  /* used in decompress_stream */
#ifdef BENCH
  unsigned long	cmsec;
  unsigned long	umsec;
  unsigned long	cbyte;
  unsigned long	ubyte;
  unsigned long	blk[2];
#endif
} file_status;


/* file_error() needs to call munmap() before it can toggle the
   EXT2_NOCOMPR_FL of the input file.  (Arguably a kernel bug.) */
static void *gMmap_start = NULL;
static size_t gMmap_len;

/* This should preferably called with gDefer_sig = true, so that we
   don't try to munmap the same area twice (if a signal is caught
   between munmap() and setting gMmap_start to NULL). */
static void
do_munmap (void)
{
  if ((gMmap_start != NULL) && (gMmap_start != ((void *) (-1))))
    {
      if (munmap (gMmap_start, gMmap_len))
	(void) fprintf (stderr, "%s: munmap failed.  Please send bug report.\n",
			prog_name);
      gMmap_start = NULL;
    }
}


/* file_error is required by catch_signal. */ 
static int file_error (file_status const * const status,
		       char const *msg1, const char *msg2, int err);

/* Set only in mf_process_file before calling mfo_open_files().  Only read
   in catch_signal(). */
static file_status *gStatus = NULL;
static bool gDefer_sig = false;
static int gPending = 0;


static void
catch_signal (int sig_num)
{
  assert (sig_num != 0);

  if (gDefer_sig)
    {
      gPending = sig_num;
      return;
    }
  else
    {
      if (gStatus)
	file_error (gStatus, "Exiting due to signal: ", strsignal (sig_num), 0);
      else
	(void) fprintf (stderr,
			"%s: Exiting due to signal: %s\n",
			prog_name, strsignal (sig_num));
      exit (1);
    }
}


/* Always returns -1. */
static int 
file_error (file_status const * const status,
	    char const *msg1,
	    const char *msg2,
	    int err)
{
  unsigned long value;

  assert (status != NULL);

  /* Defer signal handling until we've finished cleaning up our files. */
  gDefer_sig = true;

  if (msg1 != NULL)
    {
      assert (msg2 != NULL);
      fprintf (stderr, "%s: %s%s", prog_name, msg1, msg2);

      if (err != 0)
	fprintf (stderr, " (%s)", strerror (err));

      fprintf (stderr, "\n");
    }

  if (status->ifd >= 0)
    {
      /* We need to munmap() before we can toggle EXT2_NOCOMPR_FL. */
      do_munmap();

      if (status->restore_iattrs)
	{
	  assert (status->valid.iattrs);
	  value = status->iattrs;
	  if (ioctl (status->ifd, EXT2_IOC_SETFLAGS, &value))
	    (void) fprintf (stderr,
			    "%s: Warning: cannot restore attributes on `%s': %s\n",
			    prog_name, status->ipath, strerror (errno));
	}

      if (status->ifd != STDIN_FILENO)
	(void) close (status->ifd);
    }

  if (status->ofd > STDOUT_FILENO)
    {
      close (status->ofd);
      if (unlink (status->opath))
	(void) fprintf (stderr, "%s: %s: %s\n", prog_name, status->opath, strerror (errno));
    }

  gStatus = NULL;
  gDefer_sig = false;
  if (gPending)
    catch_signal (gPending);

  return -1;
}


/* Always returns 1. */
static int
fatal_error (file_status const * const status,
	     char const *msg1,
	     char const *msg2,
	     int err)
{
  file_error (status, msg1, msg2, err);
  (void) fprintf (stderr, "%s: Exiting.\n", prog_name);
  return 1;
}


static int 
mfo_open_files (file_status *status)
{
  unsigned long value;
  struct stat st;

  assert (!streq(status->ipath, "-"));

  status->restore_iattrs = false;
  status->valid.iattrs = false;
  status->valid.clu_nblocks = false;

  /* Set up .opath. */
  if (status->ofd == STDOUT_FILENO)
    status->opath = "(stdout)";
  else  
    {
      /* main() passes us just the suffix in .opath.  Make .opath point to
	 a concatenation of .ipath and the suffix. */
      static size_t opath_size = 0;
      static char *opath = NULL;
      size_t opath_len = strlen (status->ipath) + strlen (status->opath) + 1;

      if (opath_len > opath_size)
	opath = realloc (opath, opath_size = opath_len);  /* effic: Round up to multiple of 64 or so. */
      if (opath == NULL)
	{
	  (void) fprintf (stderr, "%s: failed to realloc %lu bytes for opath.  Exiting.\n",
			  prog_name, (unsigned long) opath_len);
	  exit (1);
	}
      sprintf (opath, "%s%s", status->ipath, status->opath);
      status->opath = opath;
    }

#ifdef BENCH
  status->cmsec = 0;
  status->umsec = 0;
  status->cbyte = 0;
  status->ubyte = 0;
  status->blk[0] = 0;
  status->blk[1] = 0;
#endif

  /*
   *	Open input file.
   */
  {
    if (lstat (status->ipath, &st) < 0)
      return file_error (status, status->ipath, ": couldn't stat", errno);

    if (S_ISLNK (st.st_mode))
      {
	if (status->do_continue)
	  (void) fprintf (stderr,
			  "%s: %s: warning: this is a symbolic link.  It will be replaced with a regular file, with the symlinked file untouched.\n",
			  prog_name, status->ipath);
	else
	  return file_error (status, status->ipath, ": is a symbolic link", 0);
      }
    else if (!S_ISREG (st.st_mode))
      return file_error (status, status->ipath, ": not a regular file", 0);

    if ((status->ifd = open (status->ipath, O_RDONLY, 0)) == -1)
      return file_error (status, status->ipath, ": can't open file", errno);

    if (ioctl (status->ifd, (int) EXT2_IOC_GETFLAGS, &value) == -1)
      {
	if (status->action == ACT_compress)
	  return file_error (status, status->ipath, ": can't read flags.  (Is it on an ext2 fs?)", errno);
      }
    else
      {
	status->iattrs = value;
	status->valid.iattrs = true;
      }

    if (fstat (status->ifd, &st) == -1)
      return file_error (status, status->ipath, ": can't stat file", errno);

    status->ilen = st.st_size;
    status->imode = st.st_mode;
    status->timestamps.actime = st.st_atime;
    status->timestamps.modtime = st.st_mtime;
    status->i_uid = st.st_uid;
    status->i_gid = st.st_gid;

    if (! S_ISREG (st.st_mode))
      return file_error (status, status->ipath, ": not a regular file", 0);

    if ((st.st_nlink > 1) && !status->do_continue) /* todo: warn even if --force. */
      return file_error (status, status->ipath, ": file has more than one link", 0);

    switch (status->action)
      {
      case ACT_compress:
	assert ((status->fs_blk_size == 0)
		|| (valid_blocksize (status->fs_blk_size)));

	status->clu_nchars = status->clu_nblocks * (status->fs_blk_size
						    ? status->fs_blk_size
						    : 1024);
	break;
      case ACT_decompress:
	/* If we are decompressing, we attempt to turn off automatic
	   decompression for input file, and read the cluster size.  */

	if (status->valid.iattrs
	    && ((status->iattrs & (EXT2_COMPR_FL | EXT2_COMPRBLK_FL)) == 0)
	    && !status->do_continue)
	  return file_error (status, status->ipath, ": file not compressed", 0);

	if ((status->valid.iattrs)
	    && ((status->iattrs & (EXT2_COMPR_FL | EXT2_COMPRBLK_FL)) != 0)
	    && (ioctl (status->ifd, (int) EXT2_IOC_GETCLUSTERSIZE, &value) == 0))
	  {
	    status->file_e2compr_capable = true;
	    status->clu_nblocks = value;
	    status->valid.clu_nblocks = true;
	  }
	else
	  status->file_e2compr_capable = false;

	if ((status->valid.iattrs)
	    && ((status->iattrs & (EXT2_COMPR_FL | EXT2_COMPRBLK_FL)) != 0)
	    && !(status->iattrs & EXT2_NOCOMPR_FL))
	  {
	    value = status->iattrs | EXT2_NOCOMPR_FL;
	    status->restore_iattrs = true;
	    if (ioctl (status->ifd, EXT2_IOC_SETFLAGS, &value) == -1)
	      return file_error (status, status->ipath,
				 ": can't turn off compression", errno);
	    /* TODO: at the moment we give an error for compressing if
               we're not compressing ext2 files under an e2compr
               kernel.  It would be nice to be more lenient, for
               testing purposes. */
	  }
	break;
      }
  }

  /*
   *	Set up output file.
   */
  if (status->ofd != STDOUT_FILENO) /* todo: We should still do most of
				       this even for stdout,
				       but not open() and
				       don't give up if something
				       fails. */
    {
      struct stat ost;

      if ((status->ofd = open (status->opath,
			       O_RDWR | O_CREAT | O_EXCL,
			       0600)) == -1)
	return file_error (status, status->opath,
			   ": can't open output file", errno);

      if (fstat (status->ofd, &ost) == -1)
	return file_error (status, status->opath,
			   ": can't stat temporary file", errno);
    
      if (st.st_dev != ost.st_dev)
	return file_error (status, "",
			   "input and temporary files not on same device", 0);

      if (st.st_ino == ost.st_ino)
	return file_error (status, "",
			   "input and temporary files are the same", 0);

      /* The reason we trunc to 0 before setting the size to ilen is
	 to fill it with holes. */
      if (ftruncate (status->ofd, 0) == -1)
	return file_error (status, status->opath, ":", errno);
    }

  /*
   *	Copy original properties.
   */

  switch (status->action)
    {
    case ACT_compress:
      /*
       *	If we are compressing, turn off automatic compression
       *	for the output file.
       */

      if (ioctl (status->ofd, FIGETBSZ, &value))
	return file_error (status, status->opath, ": can't read block size", errno);
      status->fs_blk_size = value;

      value = EXT2_NOCOMPR_FL | EXT2_COMPR_FL;
      if (ioctl (status->ofd, EXT2_IOC_SETFLAGS, &value) == -1)
	return file_error (status, status->opath, ": can't disable compression", errno);

      value = status->clu_nblocks;
      if (ioctl (status->ofd, EXT2_IOC_SETCLUSTERSIZE, &value) == -1)
	return file_error (status, status->opath, ": can't set cluster size", errno);

      value = status->compr_meth;
      if (ioctl (status->ofd, EXT2_IOC_SETCOMPRMETHOD, &value) == -1)
	fprintf(stderr,
		"%s: warning: kernel doesn't know about compression method `%s'\n",
		prog_name, algo[status->compr_meth].name);
      break;
    case ACT_decompress:
      value = 0;
      if (ioctl (status->ofd, EXT2_IOC_SETFLAGS, &value) == -1)
	(void) fprintf (stderr,
			"%s: Warning: can't remove attributes from `%s' (%s).\n",
			prog_name, status->opath, strerror (errno));
      /* FALL THROUGH */
    case ACT_check:
      if (status->fs_blk_size == 0)
	{
	  /* Try to read blocksize with ioctl. */
	  if (ioctl (status->ifd, FIGETBSZ, &value))
	    (void) fprintf (stderr,
			    "%s: Warning: can't read block size from `%s' (%s).\n",
			    prog_name, status->ipath, strerror (errno));
	  else
	    {
	      assert (valid_blocksize (value));
	      status->fs_blk_size = value;
	    }
	}
    }

  if ((status->ofd != STDOUT_FILENO)
      && (ftruncate (status->ofd, (size_t) status->ilen) == -1))
    return file_error (status, status->opath, ": can't resize output file", errno);

  if (status->verbose)
    {
      fprintf (stderr, "File '%s': %lu bytes\n",
	     status->ipath,
	     (unsigned long) status->ilen);
#if 0  /* todo: Maybe fix this up. */
      if (status->valid.clu_nchars)
	fprintf (stderr, ", %ld cluster%s\n",
		 round_up_div (status->ilen, status->clu_nchars),
		 ((!status->ilen || (status->ilen > status->clu_nchars))
		  ? "s"
		  : ""));
#endif
    }
  return 0;
}


#if 0 /* todo nyi */
static int
mfs_open_stdio (file_status *status)
     /* Analog of mfo_open_files but for stdin--stdout. */
{
  status->restore_iattrs = false;
  status->valid.iattrs = false;
  status->ifd = STDIN_FILENO;
  status->ofd = STDOUT_FILENO;
  status->ipath = "(stdin)";
  status->opath = "(stdout)";

#ifdef BENCH
  status->cmsec = 0;
  status->umsec = 0;
  status->cbyte = 0;
  status->ubyte = 0;
  status->blk[0] = 0;
  status->blk[1] = 0;
#endif

  status->ilen = -1;  /* fixme: check all references to ilen. */

  if (status->action == ACT_compress)
    {
      status->clu_nchars = status->clu_nblocks * status->fs_blk_size;
      status->nclusters = -1; /* fixme: check all references to nclusters. */
    }
  else
    {
      /* todo */
    }

  status->timestamps.actime = fixme;
  status->timestamps.modtime = fixme;
  return 0;
}
#endif /* stdio */


/*
 *	Everything is done, we delete the backup file and close
 *	the output file.  Original flags are restored.
 *	Timestamps (mtime and atime) from original file are applied.
 */

/* Called by process_file at end. */

static int 
close_files (file_status *status)
{

#ifdef BENCH
  if (do_bench) 
    {
      if (ioctl (status->ofd, EXT2_IOC_GETCOMPRRATIO, &(status->blk)) == -1)
	return file_error (status, status->opath, ": can't get ratio", errno);

      cblk += status->blk[0];
      ublk += status->blk[1];

      if (status->ofd != STDOUT_FILENO)
	unlink (status->opath);
    }

  else
#endif
    {
      /* Check for unbounded critical block. */
      assert (!gDefer_sig);

      /* We don't want to exit on signal after renaming but before
         restoring original properties. */
      gDefer_sig = true;

      if (status->ofd != STDOUT_FILENO)
	if (rename (status->opath, status->ipath) == -1)
	  return file_error (status, status->ipath,
			     ": can't be replaced with temp file", errno);
    }

  /* Restore timestamps, permissions, attributes, and owner/group. */
  if (status->ofd != STDOUT_FILENO)
    {
      /* impl: Note the order is important here.  Owner/group must be
	 set before permissions if suid/sgid bits are to be preserved.
	 Both must occur before setting the attributes, in case the
	 `immutable' (+i) attribute is set (which disallows setting
	 permissions or owner/group). */

      /* Owner & group. */
      {
	if (fchown (status->ofd, status->i_uid, status->i_gid))
	  fprintf (stderr,
		   "%s: Warning: couldn't restore original owner/group on `%s': %s\n",
		   prog_name, status->ipath, strerror (errno));
      }

      /* Permissions. */
      if (fchmod (status->ofd, status->imode & ~S_IFMT))
	{
	  fprintf (stderr,
		   "%s: Warning: couldn't set permissions on `%s': %s\n",
		   prog_name, status->ipath, strerror (errno));
	}

      /* Attributes. */
      if (status->valid.iattrs)
	{
	  unsigned long value = status->iattrs;

	  switch (status->action)
	    {
	    case ACT_compress:
	      value |= (unsigned long) EXT2_COMPR_FL;
	      value &= (unsigned long) ~(EXT2_NOCOMPR_FL | EXT2_ECOMPR_FL);
	      break;
	    case ACT_decompress:
	      value &= (unsigned long) ~(EXT2_COMPR_FL | EXT2_NOCOMPR_FL | EXT2_ECOMPR_FL);
	      break;
	    case ACT_check:
	      break;
	    }

	  if (ioctl (status->ofd, EXT2_IOC_SETFLAGS, &value))
	    {
	      fprintf (stderr,
		       "%s: Warning: couldn't restore attributes (0x%lx) on `%s': %s\n",
		       prog_name, status->iattrs, status->ipath, strerror (errno));
	    }
	}

      /* Timestamps. */
      if ((status->ofd != STDOUT_FILENO)
	  && (utime (status->ipath, &status->timestamps)))
	{
	  fprintf (stderr,
		   "%s: Warning: couldn't set timestamps on `%s': %s\n",
		   prog_name, status->ipath, strerror (errno));
	}
    }

  if (status->ifd >= 0)
    {
      if (status->restore_iattrs)
	{
	  unsigned long value = status->iattrs;

	  assert (status->valid.iattrs);
	  if (ioctl (status->ifd, EXT2_IOC_SETFLAGS, &value))
	    (void) fprintf (stderr,
			    "%s: Warning: cannot restore attributes on `%s': %s\n",
			    prog_name, status->ipath, strerror (errno));
	}

      if (status->ifd != STDIN_FILENO)
	(void) close (status->ifd);
    }

  if (status->ofd > STDOUT_FILENO)
    close (status->ofd);

  /* Make sure that the signal handler doesn't try to clean up too. */
  status->ifd = status->ofd = -1;

  /* OK, it's safe to exit if we've received a signal now. */
  {
    gDefer_sig = false;
    if (gPending)
      catch_signal (gPending);
  }

  return 0;
}


static int
mfc_close_io (file_status *status)
{
  int rc;

  if (status->ifd == STDIN_FILENO)
    assert (0 && "nyi");  /* todo */
  else
    rc = close_files (status);

  if (rc)
    return rc;

#ifdef BENCH
  /* Display results. */
  if (do_bench)
    {
      printf ("%-12s  %7ld  %7ld   %5.1f",
	      status->ipath, status->ubyte, status->cbyte, 
	      (status->blk[1] != 0 && status->blk[1] != status->blk[0] 
	       ? (double) 100.0 * status->blk[0] / status->blk[1] 
	       : 100.0));
    
      if (status->cmsec > 0)
	printf ("  %5.2f MB/s",
		(double) status->ubyte * status->nloop / (1000000.0 * status->cmsec));
      else 
	printf ("   ?.?? MB/s");

      if (status->umsec > 0)
	printf (" %5.2f MB/s",
		(double) status->ubyte * status->nloop / (1000000.0 * status->umsec));
      else 
	printf ("  ?.?? MB/s");

      if (status->umsec < 500 || status->cmsec < 500)
	printf ("  *\n"); 
      else 
	printf ("\n");
    }

  /* fixbe: should these 4 lines be inside the `if' block? */
  cmsec += status->cmsec;
  umsec += status->umsec;
  cbyte += status->cbyte;
  ubyte += status->ubyte;
#endif

  return 0;
}


/* todo: Make a more robust verification.  For files, try to
   decompress the whole file after it has been written.  (This should
   catch some I/O problems, and test the decompress_stream code.)  For
   stdio, at least test the decompress_stream code on every cluster
   (compressed or not). */

static inline int
mflcv_verify_compression (size_t const ulen,
			  size_t const clen,
			  file_status * const status,
			  unsigned char const * const ubuf,
			  unsigned char * const cbuf)
{
  size_t vlen;
  static unsigned char *vbuf = NULL;
#ifdef BENCH
  struct rusage old, new;
#endif

  /* effic: We can probably share ubuf or cbuf. */
  if ((vbuf = realloc (vbuf, status->clu_nchars)) == NULL)
    exit (fatal_error (status, "", "cannot allocate vbuf", 0));

#ifdef BENCH
  getrusage (RUSAGE_SELF, &old);
	  
  for (i = 0; i < status->nloop; i++)
    {
#endif
      if (algo[status->compr_meth].init (heap_base) > HEAP_SIZE)
	exit (fatal_error (status, "", "heap not big enough !!!", 0));

      vlen = algo[status->compr_meth].uncompress (cbuf + sizeof (struct ext2_cluster_head_03x),
						  vbuf,
						  clen,
						  ulen,
						  algo[status->compr_meth].xarg);

#ifdef BENCH
    }

  getrusage (RUSAGE_SELF, &new);
  status->umsec += ((new.ru_utime.tv_sec - old.ru_utime.tv_sec) * 1000000
		    + new.ru_utime.tv_usec - old.ru_utime.tv_usec);
#endif
  /* tbi: Can ulen equal zero at this point? */
  if ((vlen <= 0) || (vlen != ulen) || (memcmp (ubuf, vbuf, ulen) != 0))
    {
      (void) fatal_error (status, status->opath, ": verification failed", 0);
      abort ();
    }

  return 0;
}

static void
mflch_create_head (unsigned const ulen,
		   unsigned const clen,
		   unsigned long const ubitmap,
		   unsigned char const * const ubuf,
		   unsigned char * const cbuf,
		   int compr_meth,
		   u32 checksum)
{
  struct ext2_cluster_head_03x *head_le  = (struct ext2_cluster_head_03x *) cbuf;
  head_le->magic	= cpu_to_le16(EXT2_COMPRESS_MAGIC_03X);
  head_le->len	    	= cpu_to_le16(ulen);
  head_le->compr_len	= cpu_to_le16(clen);
  assert ((compr_meth >= 0) && (compr_meth <= 24));
  head_le->method	= compr_meth;
  head_le->reserved_0	= 0;
  head_le->bitmap	= cpu_to_le32(ubitmap);
  head_le->reserved_2	= cpu_to_le32(checksum);
}


static void *zero_buf;

static void
init_zero_buf (void)
{
  zero_buf = malloc (32*1024);
  if (zero_buf == NULL)
    {
      (void) fprintf (stderr, "%s: no memory for zero_buf\n", prog_name);
      exit(1);
    }

  /* impl: We test it for zero before memset'ing it to maximise the chance
     that we keep copy-on-write zero pages there. */
  if (!is_zero (zero_buf, 32*1024))
    memset (zero_buf, 0, 32*1024);
}


static inline int
mflc_compress_cluster (unsigned long const cluster,
		       struct file_status * const status,
		       unsigned const ulen,
		       int * const oleft_p,
		       unsigned char const ** const optr_p,
		       unsigned long const ubitmap,
		       unsigned long * const obitmap,
		       unsigned char * const ibuf,
		       unsigned char * const obuf)
{
  size_t clen, maxlen;
#ifdef BENCH
  int i;
  struct rusage old;
  struct rusage new;
#endif

  assert (valid_blocksize (status->fs_blk_size));
  if (ulen <= status->fs_blk_size)
    clen = 0;   /* Can't compress to fewer than one block. */
  else
    {
#ifdef BENCH
      getrusage (RUSAGE_SELF, &old);

      for (i = 0; i < status->nloop; i++)
	{
#endif
	  if (algo[status->compr_meth].init (heap_base) > HEAP_SIZE)
	    exit (fatal_error (status, "", "heap not big enough !!!", 0));

	  maxlen = (((ulen - 1)
		     & -status->fs_blk_size)
		    - sizeof (struct ext2_cluster_head_03x));
	  clen = algo[status->compr_meth].compress (ibuf, 
						    obuf + sizeof (struct ext2_cluster_head_03x), 
						    ulen, 
						    maxlen,
						    algo[status->compr_meth].xarg);
	  if (clen >= maxlen)
	    {
	      if (clen > EXT2_MAX_CLUSTER_BYTES - sizeof(struct ext2_cluster_head_03x))
		fprintf (stderr, "WARNING!!!");
	      else
		fprintf (stderr, "Note:");
	      fprintf (stderr, " clen=%u where maxlen=%u\n",
		       clen, maxlen);
	      clen = 0;
	    }

#ifdef BENCH
	}

      getrusage (RUSAGE_SELF, &new);

      status->cmsec += ((new.ru_utime.tv_sec - old.ru_utime.tv_sec) * 1000000
			+ new.ru_utime.tv_usec - old.ru_utime.tv_usec);
      status->ubyte += ulen;
      status->cbyte += (clen ? clen : ulen);
#endif

    }

  if (clen == 0)
    {   	/* Couldn't get it any smaller. */
      /* fixme: CLRCLUSTERBIT isn't available in e2compr-0.4.0, so make sure tht
         it definitely isn't needed.  I think this is only an issue if we're
         trying to compress in place. */
      if (ioctl (status->ofd, EXT2_IOC_CLRCLUSTERBIT, &cluster) == -1)
	return file_error (status, status->opath, ": can't clear cluster bit", errno);

      *optr_p = ibuf;
      *oleft_p = ulen;
      *obitmap = ubitmap;

      if (status->verbose)
	{
	  long nblks = round_up_div (ulen, status->fs_blk_size);

	  fprintf (stderr, 
		   " cluster #%3ld left uncompressed.  ubitmap=0x%08lx, blocks=%2ld\n",
		   cluster, ubitmap, nblks);
	}
    }
  else
    {
      unsigned long checksum;

      if (ioctl (status->ofd, EXT2_IOC_SETCLUSTERBIT, &cluster) == -1)
	{
	  return file_error (status, status->opath, ": can't set cluster bit", errno);

	  /* TODO: More graceful behaviour (considering the limited number
	     of cluster bits in current kernel implementation) would be to
	     print a warning and copy the rest of the file in its
	     uncompressed form. */
	}

      checksum = ext2_adler32 (1, ibuf, ulen);
      mflch_create_head (ulen, clen, ubitmap, ibuf, obuf, status->compr_meth, checksum);
	  
      /* todo: Better for -f would be to verify but print a warning
         if failed (rather than abort).  Or maybe have a separate flag
         that allows erroneous compression.  Also it would be nice to
         have a flag that skipped verification. */
      if (!status->do_continue
#ifdef BENCH
	  || status->do_bench
#endif
	  )
	if (mflcv_verify_compression (ulen, clen, status, ibuf, obuf))
	  return -1;

      /* Set *obitmap. */
      {
	/* Note that output block positions must be a subset of the
	   positions of the uncompressed blocks.  They must also be the first such 
	   positions available. */
	long nblks = round_up_div (clen + sizeof (struct ext2_cluster_head_03x), 
				   status->fs_blk_size);

	/* We handle the usual case first, for speed. */
	unsigned long mask = (1 << nblks) - 1;

	if ((ubitmap & mask) == mask)
	  *obitmap = mask;
	else
	  {
	    unsigned long killbits;

	    for (killbits = ubitmap; nblks--;)
	      killbits &= (killbits - 1);  /* Clear lowest set bit. */
	    assert ((killbits & ubitmap) == killbits);
	    *obitmap = ubitmap ^ killbits;
	    assert ((ubitmap & *obitmap) == *obitmap);
	  }
      }
      *optr_p = obuf;
      *oleft_p = clen + sizeof (struct ext2_cluster_head_03x);

      if (status->verbose)
	fprintf (stderr, " cluster #%3ld, ubitmap=0x%08lx, blocks=%2ld/%2ld, adler32=0x%08lx\n",
		 cluster, ubitmap, 
		 round_up_div (clen + sizeof (struct ext2_cluster_head_03x), 
			       status->fs_blk_size),
		 round_up_div (ulen, status->fs_blk_size),
		 checksum);
    }

  return 0;
}


#if 0
/* todo: This fails if fd is /dev/null. */
static void
check_offset (int fd, off_t offset)
{
  int err;

  err = lseek (fd, 0, SEEK_CUR);
  assert ((err == offset) || (err == -1));
}
#else
# define check_offset(x,y) 	/* (nothing) */
#endif


static int
write_hole (char const *zero_ptr,
	    off_t len,
	    file_status const *status)
{
  assert (len >= 0);
  if (!len)
    return 0;

  if (lseek (status->ofd, len, SEEK_CUR) < 0)
    if (writen (status->ofd, zero_ptr, len) < 0)
      return file_error (status, status->opath, ": can't write hole", errno);

  return 0;
}


static int
write_cluster( 
	       file_status const * const  status,
	       unsigned  		   oleft,
	       /* oleft: N. non-hole chars to write; length of
		  significant chars from optr.  Except for the last
		  cluster, oleft = (n. bits set in obitmap) *
		  status->fs_blk_size. */
	       long			   advance_len,
	       /* advance_len: Total amount that i_ix should advance.
		  We need this so that we can write the holes properly
		  (for stdout). */
	       char const *		   optr,     /* Pointer to cluster chars. */
	       unsigned long 		   obitmap,  /* How to distribute the chars. */
	       off_t    		   offset)   /* Where in the file to
						        start writing. */
{
  /*
   *	Write back the cluster.
   */
  long pagesize = sysconf(_SC_PAGESIZE);
  
  if (pagesize < 0)
    pagesize = PAGE_SIZE;
  /* If you have trouble compiling the above, change it to: 
     unsigned long pagesize = 8192; */

  assert (valid_blocksize (status->fs_blk_size));
  assert (is_power_of_two (pagesize));
  assert (pagesize >= status->fs_blk_size);
  assert ((offset & (status->fs_blk_size - 1)) == 0);

  if (zero_buf == NULL)
    init_zero_buf();

  /* fixme: Check that _all_ writes use write_hole() and writen(). */

  while (advance_len > 0)
    {
      /* Try to write a whole page at a time. */
      unsigned nchars_this_page = MIN (-(offset | -pagesize),
				       advance_len);
      /* pagesize - (offset % pagesize)
	 pagesize - (offset & (pagesize - 1))
	 (-(offset + 1) & (pagesize - 1)) + 1
	 ((-1 - offset) & (pagesize - 1)) + 1
	 (~offset & (pagesize - 1)) + 1
	 ~(offset | ~(pagesize - 1)) + 1
	 -1 - (offset | -pagesize) + 1
	 -(offset | -pagesize) */
      unsigned nblocks_this_page = round_up_div (nchars_this_page, status->fs_blk_size);
      u32 mask = (1 << nblocks_this_page) - 1;

      assert ((obitmap == 0) == (oleft == 0));

      switch ((obitmap + 1) & mask)
	{
	case 1: /* all holes */
	  if (write_hole (zero_buf, nchars_this_page, status))
	    return -1;

	  obitmap >>= nblocks_this_page;
	  offset += nchars_this_page;
	  advance_len -= nchars_this_page;
	  break;
	case 0: /* all non-holes */
	  if (writen (status->ofd, optr, nchars_this_page) != nchars_this_page)
	    return file_error (status, status->opath, ": can't write to output file", errno);

	  assert (oleft >= nchars_this_page);
	  oleft -= nchars_this_page;
	  optr += nchars_this_page;

	  obitmap >>= nblocks_this_page;
	  offset += nchars_this_page;
	  advance_len -= nchars_this_page;
	  break;
	default: /* mixture of holes and non-holes */
	  /* Do it one block at a time. */
	  while (nblocks_this_page--)
	    {
	      size_t write_len = MIN (advance_len, status->fs_blk_size);
	      
	      check_offset (status->ofd, offset);
	      if ((obitmap & 1) == 0)
		{
		  if (write_hole (zero_buf, write_len, status))
		    return -1;
		}
	      else
		{
		  if (writen (status->ofd, optr, write_len) != write_len)
		    return file_error (status, status->opath,
				       ": can't write to output file",
				       errno);

		  assert (oleft >= write_len);
		  oleft -= write_len;
		  optr  += write_len;
		}

	      advance_len -= write_len;
	      offset += status->fs_blk_size;
	      obitmap >>= 1;
	    }
	  break;
	}
    }
  assert (obitmap == 0);
  assert (advance_len == 0);
  assert (oleft == 0);
  return 0;
}	


enum warn_cmd {
  warn_rsrvd0, warn_long_ulen, warn_clen, warn_bitcount, warn_misaligned_ulen,
  warn_uncmp_bitmap, warn_compression_error, warn_checksum_error, warn_getbit,
  something_very_wrong, warn_not_hole, warn_eof, warn_nyi, warn_bitmap,
  warn_nolook_clusterbit, warn_believe_uncmp_bitmap, warn_cmp_bitmap,
  warn_unknown_method, warn_magic_no_holes,
  n_warnings,	/* This is not a warning but a count. */
  begin_stream, end_stream
};

static char const *
warn_str (enum warn_cmd cmd)
{
  switch (cmd)
    {
    case warn_rsrvd0:
      return "Non-zero reserved_0 field in head.  The rest of the cluster was OK and decompresses without error, so we've used the decompressed data, but the cluster may have been created with a newer version of the software, which is cause to be wary.";
    case warn_long_ulen:
      return "Uncompressed cluster len exceeds 32KB.";
    case warn_clen:
      return "Compressed data would take up at least as many blocks as uncompressed.";
    case warn_bitcount:
      return "Head bitmap seems to conflict with uncompressed length field.";
    case warn_bitmap:
      return "Head bitmap seems in error, but cluster otherwise valid.  Treated as not compressed.";
    case warn_misaligned_ulen:
      return "Length not a multiple of determined minimum block size.";
    case warn_uncmp_bitmap:
      return "Disregarded cluster bitmap indicating that this cluster is not compressed.";
    case warn_cmp_bitmap:
      return "Disregarded cluster bitmap indicating that this cluster is compressed.";
    case warn_believe_uncmp_bitmap:
      return "This cluster is a valid compressed cluster, but the cluster bitmap says that it isn't compressed.  Treated as not compressed.";
    case warn_getbit:
      return "Failed to read cluster bit.";
    case warn_unknown_method:
      return "Unknown compression method id.";
    case warn_compression_error:
      return "The head appears valid, but the data couldn't be uncompressed.";
    case warn_checksum_error:
      return "Checksum error.";
    case something_very_wrong:
      return "Cluster decompressed OK, but the information conflicts with a previous cluster.  An error has probably been made.";
    case warn_not_hole:
      return "Non-null data read where a hole was expected.";
    case warn_magic_no_holes:
      return "A magic number was found, but not at the beginning of cluster.  You can expect this to happen about once per 64MB of uncompressed clusters.  If it happens more often, you may want to check the file validity yourself.";
    case warn_eof:
      return "Unexpected end-of-file.  Leaving raw data.";
    case warn_nyi:
      return "The cluster is almost certainly compressed, but a rare circumstance has occurred that hasn't yet been coded for.  (Sorry.)  Write to the e2compr list and complain.";
    case warn_nolook_clusterbit:
      return "Couldn't examine cluster bitmap.";
    case n_warnings:
    case begin_stream:
    case end_stream:
      assert (0 && "unexpected warning");
      abort();
    }
  assert (0 && "can't reach here");
  return NULL;
}

static void
warn (enum warn_cmd cmd, char const *id)
{
  static int warn_count[n_warnings];

  switch (cmd)
    {
    case begin_stream:
      (void) memset (warn_count, 0, sizeof (*warn_count) * n_warnings);
      return;
    case end_stream:
      {
	int i;
	bool head_printed = false;

	for (i = 0; i < n_warnings; i++)
	  if (warn_count[i] != 0)
	    {
	      if (!head_printed)
		{
		  (void) fprintf (stderr,
				  "%s: The following warning(s) "
				  "were issued while decompressing %s:\n",
				  prog_name,
				  id);
		  head_printed = true;
		}
	      (void) fprintf (stderr, "  %s\n", warn_str (i));
	      /*
	      (void) fprintf (stderr, "    (cluster%s at byte offset%s");
	      for j in offset associated with warning #i:
	      (void) fprintf (stderr, " %dK", j / 1024);
	      // Or better yet, put some code in to recognise linear progressions.
	      */
	      if (warn_count[i] > 1)
		(void) fprintf (stderr, "  (Last message occurs %d times.)\n", warn_count[i]);
	    }
      }
      return;
    default:
      warn_count[cmd]++;
      return;
    }
  assert (0 && "Can't reach here.");
}


#if 0  /* Not used. */
static int
min (int a, int b)
{
  return (a < b
	  ? a
	  : b);
}



static int
write_holey_rest (int fd, char const *buf, int ntogo, int fsblk_nchars)
{
  int n_write, n_skip;
  
  assert (ntogo >= 0);
  while (ntogo > 0)
    {
      for (n_write = 0;
	   (n_write < ntogo) && (!is_zero (buf + n_write, (size_t) fsblk_nchars));
	   n_write += fsblk_nchars)
	;
      n_write = min (ntogo, n_write);
      ntogo -= n_write;
      buf += n_write;
      if (writen (fd, buf, n_write) < 0)
	return -1;

      for (n_skip = 0;
	   (n_skip < ntogo) && (is_zero (buf + n_skip, fsblk_nchars));
	   n_skip += fsblk_nchars)
	;
      n_skip = min (ntogo, n_skip);
      ntogo -= n_skip;
      buf += n_skip;
      if (lseek (fd, n_skip, SEEK_CUR) < 0)
	return -1;

      assert (ntogo >= 0);
    }
  return 0;
}


# if 1 /* Efficient. */

static int
gcd (register int a0, register int b0)
{
  register int a1, b1;

  assert (a0 >= 0);
  assert (b0 > 0);

  if (a0 == 0)
    return b0;

  for (;;)
    {
      if ((a1 = (b0 % (b1 = a0)))
	  == 0)
	return b1;

      if ((a0 = (b1 % (b0 = a1)))
	  == 0)
	return b0;
    }
}

# else /* More readable equivalent. */

static int
gcd (int a, int b)
{
  assert (a >= 0);
  assert (b > 0);

  if (a == 0)
    return b;
  return gcd (b % a, a);
}

# endif /* Which gcd(). */
#endif /* Not used. */


static int
decompress_stream (file_status * const status,
		   const char * const ichars,   /* Pointer to mmap'ed file. */
		   char * const cbuf,
		   char * const obuf)
{
  unsigned clu_min_nchars, clu_max_nchars;
  unsigned fsblk_min_nchars, fsblk_max_nchars;
  off_t i_ix;

  if (status->fs_blk_size == 0)
    { /* Only happens if the FIGETBSZ ioctl failed. */
      fsblk_min_nchars = EXT2_MIN_BLOCK_SIZE;
      fsblk_max_nchars = EXT2_MAX_BLOCK_SIZE;
    }
  else
    fsblk_min_nchars = fsblk_max_nchars = status->fs_blk_size;

  /* Determine starting minimum and maximum cluster size. */
  {
    if (status->valid.clu_nblocks && !status->stress_test)
      {
	clu_min_nchars = status->clu_nblocks * fsblk_min_nchars;
	clu_max_nchars = MIN (status->clu_nblocks * fsblk_max_nchars,
			      EXT2_MAX_CLUSTER_BYTES);
	assert (clu_min_nchars >= 4096);
	assert (clu_max_nchars <= EXT2_MAX_CLUSTER_BYTES);
	assert (clu_min_nchars <= clu_max_nchars);
      }
    else
      {
	clu_min_nchars = 4096;
	clu_max_nchars = EXT2_MAX_CLUSTER_BYTES;
      }
  }

  /* N.B. In the following I've allowed for the possibility that a
     cluster that is compressed is returned to us uncompressed.  This
     possibility is the result of a bug in earlier e2compr kernel
     code, where the NOCOMPR flag was not honoured for reads.  (This
     bug is fixed in recent 0.3 versions of e2compr.)

     So we have the following cases:

     i) No e2compr support.

     ii) Has e2compr support, but the NOCOMPR flag never honoured for
     reading.  This is the state of the publically-released code.

     iii) Has e2compr support and reads with NOCOMPR are correctly
     implemented. */

  warn (begin_stream, status->ipath);

  /* Main loop. */
  for (i_ix = 0; i_ix < status->ilen;)
    {
      unsigned provis_fsblk_min_nchars;  /* effic: Maybe keep the
					    log2, so we can use shift. */
      unsigned provis_fsblk_max_nchars;
      unsigned provis_clu_min_nchars;
      unsigned provis_clu_max_nchars;
      struct ext2_cluster_head_03x const *head_le;
      /* head_le points to the cluster head on disk (which is in
         little-endian format). */
      off_t head_offset; /* == ((char const *) head_le) - ichars. */
      struct {
	unsigned len;
	unsigned compr_len;
	int method;
	u32 bitmap;
	u32 checksum;
      } head;
      char const *earliest_clu_begin;
      /* Until we know the cluster size, we don't know which hole
	 blocks belong with the cluster head and which belong to the
	 previous cluster/s.  All blocks between
	 earliest_clu_begin and the head are holes.
	 earliest_clu_begin is on a cluster boundary
	 (provis_clu_min_nchars). */
      char const *latest_clu_end;
      unsigned nbits;
      /* nbits is the number of `one' bits in head.bitmap. */
      char const *deholed_ptr;
      /* deholed_ptr points to an area where the compressed data is stored
	 contiguously (i.e. without intervening hole blocks). */
      char const *hole_begin;
      /* hole_begin eventually points to the first block after the
	 last block occupied by compressed data.  From hole_begin to
	 the end of the cluster should all be holes in a valid
	 compressed cluster. */
      int n_hole_bits;
      /* n_hole_bits is the count of zero bits at the beginning of
	 head.bitmap, i.e. the number of hole blocks that there should
	 be between the start of the cluster and the head. */
      u32 trunc_bitmap;
      /* trunc_bitmap is head.bitmap with all low `zero' bits shifted out. */


      /* Consider the cluster to be compressed until we find otherwise. */

      
      if (status->stress_test)
	{
	  /* Forget what we've learnt from previous clusters. */
	  fsblk_min_nchars = EXT2_MIN_BLOCK_SIZE;
	  fsblk_max_nchars = EXT2_MAX_BLOCK_SIZE;
	  clu_min_nchars = 4096;
	  clu_max_nchars = EXT2_MAX_CLUSTER_BYTES;
	}

      /* Initialisation. */
      {
	earliest_clu_begin = ichars + i_ix;
	latest_clu_end = ichars + MIN(status->ilen,
				      (i_ix | (clu_max_nchars - 1)) + 1);
	deholed_ptr = NULL;
	hole_begin = NULL;
	n_hole_bits = -1;
      }

      /* Find head.  (Needn't be first block, but must be first
	 non-hole head.) */
      {
	char const *tmp;

	assert (latest_clu_end != NULL);
	assert (earliest_clu_begin != NULL);
	assert (latest_clu_end > earliest_clu_begin);
	assert (latest_clu_end - earliest_clu_begin <= clu_max_nchars);

	head_le = NULL;
	for (tmp = earliest_clu_begin;
	     tmp + sizeof (struct ext2_cluster_head_03x) < latest_clu_end;
	     tmp += fsblk_min_nchars)
	  {
	    if (*((u16 const *) tmp) == cpu_to_le16(EXT2_COMPRESS_MAGIC))
	      {
		head_le = (struct ext2_cluster_head_03x const *) tmp;
		head_offset = tmp - ichars;
		break;
	      }
	    if (*((u16 const *) tmp) != 0)
	      goto consider_uncompressed;
	  }
	if (!head_le)
	  goto consider_uncompressed;
      }

      assert (head_le->magic == cpu_to_le16(EXT2_COMPRESS_MAGIC));

      provis_fsblk_min_nchars = fsblk_min_nchars;
      provis_fsblk_max_nchars = fsblk_max_nchars;
      provis_clu_min_nchars = clu_min_nchars;
      provis_clu_max_nchars = clu_max_nchars;

      /* Only holes can preceed the cluster head. */
      if (((char const *) head_le) > earliest_clu_begin)
	{
	  /* todo: Provide alternative based on FIBMAP ioctl.  It's
             both faster and more correct (where we can use it). */
	  if (!is_zero (earliest_clu_begin,
			((char const *) head_le) - earliest_clu_begin))
	    {
	      warn (warn_magic_no_holes, status->ipath);
	      goto consider_uncompressed;
	    }
	}

      /* We might as well write the hole now.  This just saves some
	 CPU time if it turns out not to be a compressed cluster
	 (despite having a valid magic number). */
      switch (status->action)
	{
	case ACT_decompress:
	  if (write_hole (earliest_clu_begin,
			  ((char const *) head_le) - earliest_clu_begin,
			  status))
	    return -1;

	  i_ix += ((char const *) head_le) - earliest_clu_begin;
	  break;
	case ACT_check:
	  break;
	default:
	  assert(0 && "durchfall");
	}

      /* Check our expectations on ext2_cluster_head_03x.  This should
         compile to nothing. */
      {
	assert ((((char const *) (&head_le->magic))
		 - ((char const *) head_le)) == 0);
	assert ((((char const *) (&head_le->len))
		 - ((char const *) head_le)) == 2);
	assert ((((char const *) (&head_le->compr_len))
		 - ((char const *) head_le)) == 4);
	assert ((((char const *) (&head_le->method))
		 - ((char const *) head_le)) == 6);
	assert ((((char const *) (&head_le->reserved_0))
		 - ((char const *) head_le)) == 7);
	assert ((((char const *) (&head_le->bitmap))
		 - ((char const *) head_le)) == 8);
	assert ((((char const *) (&head_le->reserved_2))
		 - ((char const *) head_le)) == 12);
	assert (sizeof (struct ext2_cluster_head_03x) == 16);

	assert (sizeof (head_le->len) * BITSPERBYTE == 16);
	assert (sizeof (head_le->compr_len) * BITSPERBYTE == 16);
	assert (sizeof (head_le->method) == 1);
	assert (sizeof (head_le->bitmap) * BITSPERBYTE == 32);
	assert (sizeof (head_le->reserved_2) * BITSPERBYTE == 32);
      }

      head.len   	= le16_to_cpu (head_le->len);
      head.compr_len 	= le16_to_cpu (head_le->compr_len);
      head.method 	= head_le->method;
      head.bitmap 	= le32_to_cpu (head_le->bitmap);
      head.checksum	= le32_to_cpu (head_le->reserved_2);
      
      if (head.len > EXT2_MAX_CLUSTER_BYTES)
	{
	  warn (warn_long_ulen, status->ipath);
	  goto consider_uncompressed;
	}

      /* todo: We should loop until provis_* stabilise.  For example,
	 if fsblk_nchars is more than 1024 then clu_min_nchars may
	 still be too small by the time we get to write the cluster. */

      /* Make simple adjustments to provis_fsblk_max_nchars based on bitmap. */
      {
	assert (EXT2_MAX_CLUSTER_BLOCKS == 32);
	assert (EXT2_MAX_CLUSTER_BYTES == (32 * 1024));
	assert (valid_clu_nchars (provis_clu_max_nchars,
				  provis_fsblk_min_nchars));

	/* If a bit in the top 16 bits is set then clu_nblocks must be more
	   than 16.  The only valid cluster size more than 16 is 32. */
	if (head.bitmap & ~0xffff)
	  {
	    provis_fsblk_max_nchars = provis_clu_max_nchars / 32;
	    provis_clu_min_nchars = MAX (provis_clu_min_nchars,
					 provis_fsblk_min_nchars * 32);
	  }
	else if (head.bitmap & 0xff00)
	  {
	    provis_fsblk_max_nchars = MIN (provis_fsblk_max_nchars,
					   provis_clu_max_nchars / 16);
	    provis_clu_min_nchars = MAX (provis_clu_min_nchars,
					 provis_fsblk_min_nchars * 16);
	  }
	else if (head.bitmap & 0xf0)
	  {
	    provis_fsblk_max_nchars = MIN (provis_fsblk_max_nchars,
					   provis_clu_max_nchars / 8);
	    provis_clu_min_nchars = MAX (provis_clu_min_nchars,
					 provis_fsblk_min_nchars * 8);
	  }
	else
	  {
	    /* The smallest valid cluster size is 4 blocks. */
	    provis_fsblk_max_nchars = MIN (provis_fsblk_max_nchars,
					   provis_clu_max_nchars / 4);
	    provis_clu_min_nchars = MAX (provis_clu_min_nchars,
					 provis_fsblk_min_nchars * 4);
	  }

	if ((provis_fsblk_max_nchars < provis_fsblk_min_nchars)
	    || (provis_clu_max_nchars < provis_clu_min_nchars))
	  {
	    warn (warn_bitmap, status->ipath);
	    goto consider_uncompressed;
	  }
      }

      if (round_up_ge (head.compr_len + sizeof (struct ext2_cluster_head_03x),
		       head.len,
		       provis_fsblk_min_nchars))
	{
	  /* Neither the kernel nor e2compress(1) allow compressed data to take
	     up as much space as the uncompressed data. */
	  warn (warn_clen, status->ipath);
	  goto consider_uncompressed;
	}

      /* Except for the last cluster, ,the number of bits set in
         bitmap` * fs_blksize should equal ,uncompressed len`. */
      if (head.bitmap == 0)
	{
	  warn (warn_bitcount, status->ipath);
	  goto consider_uncompressed;
	}

      nbits = count_bits (head.bitmap);

      /* todo: We could write `head.len' instead.  (However, I
	 currently have a good test file that happens to be shorter
	 than it should be, so I won't change this.  Also, the warning
	 might have to be different.) */
      if (head.compr_len + sizeof (struct ext2_cluster_head_03x)
	  > (latest_clu_end - ((char const *) head_le)))
	{
	  warn (warn_eof, status->ipath);
	  goto consider_uncompressed;
	}

      if ((latest_clu_end - ichars) & (provis_clu_max_nchars - 1))
	{
	    /* We're near EOF, so we can't rely on the length of
               uncompressed data being the same as nbits *
               blocksize. */

	  if ((long) head.len > (long) provis_fsblk_max_nchars * nbits)
	    {
	      warn (warn_bitcount, status->ipath);
	      goto consider_uncompressed;
	    }

	  /* Calculate provisional min,max block size.  (Provisional
	     in the sense that it is calculated on the assumption that
	     this is a valid compressed cluster, which we may later
	     decide not to be true. */
	  {
	    for (;; provis_fsblk_min_nchars *= 2)
	      {
		if (provis_fsblk_min_nchars > provis_fsblk_max_nchars)
		  {
		    warn (warn_bitcount, status->ipath);
		    goto consider_uncompressed;
		  }
		if (round_up_div (head.len, provis_fsblk_min_nchars) != nbits)
		  continue;
		if (round_up_div (head.compr_len + sizeof (struct ext2_cluster_head_03x),
				  provis_fsblk_min_nchars)
		    >= nbits)
		  continue;
		break;
	      }

	    for (;; provis_fsblk_max_nchars /= 2)
	      {
		if (provis_fsblk_min_nchars > provis_fsblk_max_nchars)
		  {
		    warn (warn_clen, status->ipath);
		    goto consider_uncompressed;
		  }
		if (round_up_ge (head.compr_len + sizeof (struct ext2_cluster_head_03x),
				 head.len,
				 provis_fsblk_max_nchars))
		  continue;
		break;
	      }
	  }
	}
      else
	{
	  unsigned tmp_nchars;

	  if (head.len & (provis_fsblk_min_nchars - 1))  /* assm3 */
	    {
	      warn (warn_misaligned_ulen, status->ipath);
	      goto consider_uncompressed;
	    }

	  if (head.len % nbits != 0)  /* assm3 */
	    {
	      warn (warn_misaligned_ulen, status->ipath);
	      goto consider_uncompressed;
	    }

	  tmp_nchars = head.len / nbits;
	  if ((!is_power_of_two (tmp_nchars)) /* assm1 */
	      || (tmp_nchars < provis_fsblk_min_nchars)
	      || (tmp_nchars > provis_fsblk_max_nchars))
	    {
	      warn (warn_bitcount, status->ipath);
	      goto consider_uncompressed;
	    }
	  provis_fsblk_max_nchars
	    = provis_fsblk_min_nchars
	    = tmp_nchars;
	}

      if ((head.method > highest_known_method_id)
	  || (algo[head.method].name == NULL))
	{
	  /* todo: give a number. */
	  warn (warn_unknown_method, status->ipath);
	  goto consider_uncompressed;
	}

      if (algo[head.method].init (heap_base) > HEAP_SIZE)
	exit (fatal_error (status, "", "heap not big enough !!!", 0));

      /* Set up trunc_bitmap and n_hole_bits. */
      {
	trunc_bitmap = head.bitmap;
	n_hole_bits = 0;
	while ((trunc_bitmap & 1) == 0)
	  {
	    n_hole_bits++;
	    trunc_bitmap >>= 1;
	  }
      }

      /* Set up deholed_ptr for the decompression routine.  If the input data
	 contains no holes then we can just point into ichars;
	 otherwise we have to copy the non-hole bits into contiguous
	 space in cbuf. */
      {
	unsigned provis_clu_max_nblocks
	  = MIN (provis_clu_max_nchars / provis_fsblk_min_nchars,
		 EXT2_MAX_CLUSTER_BLOCKS);

	deholed_ptr = NULL;	/* Trap bugs. */

	if (trunc_bitmap == (1 << provis_clu_max_nblocks) - 1) /* Most common case. */
	  {
	    deholed_ptr = (char const *) head_le;
	    hole_begin = (deholed_ptr
			  + round_up ((sizeof (struct ext2_cluster_head_03x)
				       + head.compr_len),
				      provis_fsblk_max_nchars));
	  }
	else
	  {
	    register u32 requiredMask;

	    /* */
	    {
	      register unsigned n;
	      /* Maximum number of contiguous blocks required at beginning of cluster. */

	      n = round_up_div (head.compr_len + sizeof (struct ext2_cluster_head_03x),
				provis_fsblk_min_nchars);
	      requiredMask = (1 << n) - 1;
	    }

	    if (((trunc_bitmap + 1) & requiredMask) == 0)
	      {
		deholed_ptr = (char const *) head_le;
		hole_begin = (deholed_ptr
			      + round_up ((sizeof (struct ext2_cluster_head_03x)
					   + head.compr_len),
					  provis_fsblk_max_nchars));
	      }
	    else if (provis_fsblk_min_nchars == provis_fsblk_max_nchars)
	      {
		char *cbuf_ptr = cbuf;
		u32 shortmap = trunc_bitmap;

		hole_begin = (char const *) head_le;
		while (shortmap != 0)
		  {
		    if ((shortmap & 1) != 0)
		      {
			memcpy (cbuf_ptr,
				hole_begin,
				provis_fsblk_min_nchars);
			cbuf_ptr += provis_fsblk_min_nchars;
		      }
		    hole_begin += provis_fsblk_min_nchars;
		    shortmap >>= 1;
		  }
		deholed_ptr = cbuf;
	      }
	    else
	      {
		/* tbi: nyi. */
		warn (warn_nyi, status->ipath);
		goto consider_uncompressed;
	      }
	  }
      }

      if (algo[head.method].uncompress (deholed_ptr + sizeof (struct ext2_cluster_head_03x),
					obuf,
					head.compr_len,
					head.len,
					algo[head.method].xarg)
	  != head.len)
	{
	  warn (warn_compression_error, status->ipath);
	  goto consider_uncompressed;
	}

#if 0  /* todo: nyi.  Make warn() accept third arg.  Search for
	  `something_very_wrong'. */
      warn (warn_decompress_succeeded, status->ipath, head_offset);
#endif

      if ((head.checksum != 0)
	  && (ext2_adler32 (1, obuf, head.len) != head.checksum))
	{
	  warn (warn_checksum_error, status->ipath);
	  goto consider_uncompressed;
	}

      if (status->file_e2compr_capable)
	{
	  unsigned long value;
	  /* N.B. Much of this is duplicated in consider_uncompressed. */

	  if ((head_offset <= provis_clu_min_nchars)
	      || (provis_clu_min_nchars == provis_clu_max_nchars))
	    {
	      value = head_offset / provis_clu_min_nchars;
	      if (ioctl (status->ifd, (int) EXT2_IOC_GETCLUSTERBIT, &value))
		warn (warn_getbit, status->ipath);
	      else if (!value)
		{
		  if (status->believe_bitmap)
		    {
		      warn (warn_believe_uncmp_bitmap, status->ipath);
		      goto consider_uncompressed;
		    }
		  else
		    warn (warn_uncmp_bitmap, status->ipath);
		}
	    }
	  else
	    {
	      /* todo: We could check all possible cluster numbers
		 and see if the result is the same for each.
		 This is quite likely to succeed. */
	      warn (warn_nolook_clusterbit, status->ipath);
	    }
	}

      if (head_le->reserved_0 != 0)
	{
	  warn (warn_rsrvd0, status->ipath);
	}

      /* Work out where the real cluster boundary is. */
      {
	int slack_len, min_slack_before, max_slack_after;
	/* slack_len == min_slack_before + max_slack_after. */

	assert (n_hole_bits >= 0);
	assert (n_hole_bits + nbits <= 32);

	slack_len = ((char const *) head_le) - earliest_clu_begin;
	assert (slack_len >= 0);

	if (n_hole_bits * provis_fsblk_min_nchars > slack_len)
	  {
	    warn (something_very_wrong, status->ipath);
	    /* todo: This may be the wrong warning. */
	    goto consider_uncompressed;
	  }

	/* Adjust provis_fsblk_max_nchars. */
	while (n_hole_bits * provis_fsblk_max_nchars > slack_len)
	  {
	    provis_fsblk_max_nchars >>= 1;
	    if (provis_fsblk_max_nchars < provis_fsblk_min_nchars)
	      {
		warn (warn_bitmap, status->ipath);
		goto consider_uncompressed;
	      }
	  }

	assert (((earliest_clu_begin - ichars)
		 & (clu_min_nchars - 1))
		== 0);

	for(;;)
	  {
	    max_slack_after = n_hole_bits * provis_fsblk_max_nchars;
	    min_slack_before = slack_len - max_slack_after;
	    if (min_slack_before & (clu_min_nchars - 1))
	      {
		provis_fsblk_max_nchars >>= 1;
		if (provis_fsblk_max_nchars < provis_fsblk_min_nchars)
		  {
		    warn (warn_bitmap, status->ipath);
		    goto consider_uncompressed;
		  }
	      }
	    else
	      break;
	  }
      }

      /* Update provis_clu_{min,max}_nchars. */
      {
	/* clu_max_nchars */
	/* We don't know the actual beginning of the cluster,
	   so we take the most conservative of all possible
	   cluster beginnings. */
	/* While there's not a ,,provis_clu_max_nchars`-sized cluster`
	   boundary between earliest_clu_begin (inclusive) and the
	   head, halve provis_clu_max_nchars. */
	while (((earliest_clu_begin - ichars - 1) & -provis_clu_max_nchars)
	       == (head_offset & -provis_clu_max_nchars))
	  {
	    provis_clu_max_nchars >>= 1;
	  }

	/* clu_min_nchars */
	while (head.len > provis_clu_min_nchars)
	  {
	    /* Ignore this test; it should compile to nothing. */
	    if ((sizeof (provis_clu_min_nchars) == 2)
		&& (provis_clu_min_nchars == provis_clu_max_nchars))
	      break;

	    provis_clu_min_nchars <<= 1;
	  }

	/* Check. */
	if (provis_clu_min_nchars > provis_clu_max_nchars)
	  {
	    warn (something_very_wrong, status->ipath);
	    goto consider_uncompressed;
	  }
      }

      /* Update provis_fsblk_min_nchars and provis_fsblk_max_nchars. */
      {
	/* Except for the last cluster, uncompressed len should equal
	   the number of bits set in bitmap multiplied by the filesystem
	   block size.  In the last cluster, this product equals the
	   rounded up uncompressed len. */
	assert (head.bitmap != 0);
	assert (valid_blocksizes (provis_fsblk_min_nchars, provis_fsblk_max_nchars));

	if ((latest_clu_end - ichars) & (provis_clu_max_nchars - 1))
	  {
	    /* We can't find an exact value, but we can at least narrow down the
	       upper limit. */
	    assert (head.compr_len + sizeof (struct ext2_cluster_head_03x) < head.len);
	    assert (!round_up_ge (head.compr_len + sizeof (struct ext2_cluster_head_03x),
				  head.len,
				  provis_fsblk_min_nchars));

	    /* ulen <= nbits * fsblk_nchars              (1)
	       ulen > (nbits - 1) * fsblk_nchars         (2)
	       nbits >= 1                               (3)
	       fsblk_nchars & (fsblk_nchars - 1) == 0   (4)
	       (1) /\ (3) ==>  fsblk_nchars >= ulen / nbits       (5)
	       (2) /\ (3) ==>  fsblk_nchars < ulen / (nbits - 1)  (6)
	       (5) ==>  fsblk_nchars >= floor (ulen / nbits)
	       (6) ==>  fsblk_nchars < ceil (ulen / (nbits - 1)) */

	    while (round_up_ge (head.compr_len + sizeof (struct ext2_cluster_head_03x),
				head.len,
				provis_fsblk_max_nchars))
	      provis_fsblk_max_nchars >>= 1;
	    if (nbits > 1)
	      {
		int tmp = (head.len - 1) / (nbits - 1);
		/* len > (nbits - 1) * fsblk_nchars  	         (1)
		   (1) <==> fsblk_nchars < len / (nbits - 1)     (2)
		   (2) ==> fsblk_nchars <= len div (nbits - 1)   (3)
		   (3) is true but could be tighter.
		   (1) /\ isint(len) /\ isint((nbits - 1) * fsblk_nchars)
		   ==> len - 1 >= (nbits - 1) * fsblk_nchars     (4)
		   (4) ==> (len - 1) / (nbits - 1) >= fsblk_nchars  (5)
		   (5) ==> fsblk_nchars =< tmp
		 */
		if (provis_fsblk_min_nchars > tmp)
		  {
		    warn (something_very_wrong, status->ipath);
		    goto consider_uncompressed;
		  }

		while (provis_fsblk_max_nchars > tmp)
		  provis_fsblk_max_nchars /= 2;
	      }

	    /* */
	    {
	      unsigned tmp = head.len / nbits;

	      if (provis_fsblk_max_nchars < tmp)
		{
		  warn (something_very_wrong, status->ipath);
		  goto consider_uncompressed;
		}

	      while (provis_fsblk_min_nchars < tmp)
		provis_fsblk_min_nchars *= 2;
	    }
	  }
	else
	  {
	    unsigned tmp_nchars;

	    tmp_nchars = head.len / nbits;
	    if ((tmp_nchars & (tmp_nchars - 1))
		|| (tmp_nchars < provis_fsblk_min_nchars)
		|| (tmp_nchars > provis_fsblk_max_nchars))
	      {
		warn (warn_bitcount, status->ipath);
		goto consider_uncompressed;
	      }

	    if (round_up_ge (head.compr_len + sizeof (struct ext2_cluster_head_03x),
			     head.len,
			     tmp_nchars))
	      {
		warn (something_very_wrong, status->ipath);
		goto consider_uncompressed;
	      }

	    /* The minimum cluster size is 4 filesystem blocks. */
	    if (4 * tmp_nchars > provis_clu_max_nchars)
	      {
		warn (something_very_wrong, status->ipath);
		goto consider_uncompressed;
	      }

	    provis_fsblk_min_nchars
	      = provis_fsblk_max_nchars
	      = tmp_nchars;
	  }      
      }

      /* todo: the above block should probably be merged with the calculations
	 that took place before decompressing. */

      /* Check that remaining blocks in cluster are zeroed (i.e. holes). */
      {
	/* todo: Use ibitmap. */
	long zero_len;

	assert ((hole_begin != NULL) && (hole_begin > (char const *) head_le));
	zero_len = (MIN (latest_clu_end,
			 ((char const *) head_le) + provis_clu_min_nchars)
		    - hole_begin);

	/* impl: It's still possible for zero_len to be less than 4 if the file
	   size is n bytes more than a multiple of the filesystem blocksize,
	   where 1 <= n < 4. */
	assert (sizeof (*hole_begin) == 1);
	if ((zero_len > 0)
	    && !is_zero (hole_begin, zero_len))
	  {
	    warn (warn_not_hole, status->ipath);
	    goto consider_uncompressed;
	    /* I'm not sure about this, but the idea is that we don't
	       lose the data kept in what's supposed to be a hole.
	       This should almost certainly be reported, preferably
	       with offset. */
	  }
      }

      if (status->verbose)
	{
	  fprintf (stderr, " head offset %5ld, obitmap=0x%08lx, blocks=%2u/%2u, adler32=0x%08lx\n",
		   (long) head_offset,
		   (unsigned long) head.bitmap, 
		   (unsigned) round_up_div ((head.compr_len 
					     + sizeof (struct ext2_cluster_head_03x)), 
					    provis_fsblk_min_nchars),
		   (unsigned) round_up_div (head.len, provis_fsblk_min_nchars),
		   (unsigned long) head.checksum);

	  /* tbi: Is provis_fsblk_min_nchars guaranteed to equal
	     provis_fsblk_max_nchars?  If this is the first and last cluster,
	     and we haven't got the blocksize from an ioctl or the
	     command line then I think not; so should we give some
	     indication of the unsurety? */
	}

      {
	off_t new_i_ix = MIN (((i_ix + 1) | (provis_clu_min_nchars - 1)) + 1,
			      latest_clu_end - ichars);

	assert (new_i_ix != i_ix);
	assert (valid_clu_nchars (provis_clu_min_nchars, provis_fsblk_min_nchars));
	assert ((trunc_bitmap & 1) != 0);

	status->fs_blk_size = provis_fsblk_min_nchars;
	if (write_cluster (status, head.len, new_i_ix - i_ix, obuf, trunc_bitmap, i_ix))
	  return -1;
	i_ix = new_i_ix;
      }

      fsblk_min_nchars = provis_fsblk_min_nchars;
      fsblk_max_nchars = provis_fsblk_max_nchars;
      clu_min_nchars = provis_clu_min_nchars;
      clu_max_nchars = provis_clu_max_nchars;
      
      goto next_block;

    consider_uncompressed:
      {
	/* Cluster is already stored in uncompressed form: copy straight to output. */
	size_t nchars;
	unsigned long bitmap = 0; /* Used in display. */
	unsigned long cluster_offset = i_ix; /* Used in display. */

	nchars = (MIN (status->ilen,
		       round_up (i_ix + 1,
				 clu_min_nchars))
		  - i_ix);

	/* Check cluster bit. */
	if (status->file_e2compr_capable)
	  {
	    unsigned long value;
	    /* N.B. Much of this is duplicated in consider_uncompressed. */

	    if ((i_ix <= clu_min_nchars)
		|| (clu_min_nchars == clu_max_nchars))
	      {
		value = i_ix / clu_min_nchars;
		if (ioctl (status->ifd, (int) EXT2_IOC_GETCLUSTERBIT, &value))
		  warn (warn_getbit, status->ipath);
		else if (value)
		  warn (warn_cmp_bitmap, status->ipath);
	      }
	    else
	      {
		/* todo: We could check all possible cluster numbers
		   and see if the result is the same for each.
		   This is quite likely to succeed. */
		warn (warn_nolook_clusterbit, status->ipath);
	      }
	  }

	/* Write it out, handling holes. */
	{
	  unsigned long bit = 1;
	  unsigned oleft = nchars;

	  assert (nchars <= EXT2_MAX_CLUSTER_BYTES);
	  assert (nchars / EXT2_MAX_CLUSTER_BLOCKS <= fsblk_min_nchars);
	  while (oleft)
	    {
	      unsigned write_len = MIN (oleft, fsblk_min_nchars);

	      check_offset(status->ofd, i_ix);
	      if (is_zero (ichars + i_ix, write_len))
		{
		  if (write_hole (ichars + i_ix, write_len, status))
		    return -1;
		}
	      else
		{
		  if (writen (status->ofd, ichars + i_ix, write_len) != write_len)
		    return file_error (status, status->opath,
				       ": can't write to output file", errno);

		  bitmap |= bit;
		}
	      oleft -= write_len;
	      i_ix += write_len;
	      bit <<= 1;
	    }
	}

	if (status->verbose)
	  {
	    if (fsblk_min_nchars == fsblk_max_nchars)
	      fprintf (stderr,
		       " cluster at offset %lu wasn't compressed.  "
		       "bitmap=%8lx.\n",
		       cluster_offset, bitmap);
	    else
	      fprintf (stderr,
		       " cluster at offset %lu wasn't compressed.  "
		       "bitmap=%8lx if fs blksize = %d.\n",
		       cluster_offset, bitmap, fsblk_min_nchars);
	  }
      }

    next_block:
      ;
    } /* for */
  warn (end_stream, status->ipath);
  return 0;
}


static inline int
mfl_process_cluster (unsigned long const cluster,
		     struct file_status * const status,
		     unsigned char * const ibuf,
		     unsigned char * const obuf,
		     off_t *ileftp,
		     off_t *offsetp)
{
  unsigned oleft;
  unsigned char const *optr;
  unsigned char *iptr = ibuf;
  unsigned long ibitmap = 0;
  unsigned long obitmap;
  unsigned ilen = 0;
  size_t tot_adv_len = MIN (*ileftp,
			    status->clu_nchars);
  /*
   *	Read the cluster, creating ibitmap as we go.
   */
  {
    int block;

    for (block = 0;
	 (block < status->clu_nblocks) && (*ileftp > 0);
	 block++)
      {
	unsigned chars = MIN (*ileftp, status->fs_blk_size);

	iptr[0] = iptr[1] = iptr[2] = iptr[3] = 0;
	if (read (status->ifd, iptr, chars) != chars)
	  return file_error (status, status->ipath, ": can't read from input file", errno);

	if ((iptr[0] != 0)
	    || (iptr[1] != 0)
	    || (iptr[2] != 0)
	    || (iptr[3] != 0)
	    || ((chars > 4)
		&& (memcmp (iptr, iptr + 4, chars - 4) != 0)))
	  {
	    ibitmap |= 1 << block;
	    iptr += chars;
	    ilen += chars;
	  }

	*ileftp -= chars;
      }
  }

  assert (status->action == ACT_compress);
  if (mflc_compress_cluster (cluster, status, ilen, &oleft, &optr, 
			     ibitmap, &obitmap, ibuf, obuf))
    return -1;

  /* */
  {
    unsigned nBlksReqd = round_up_div (oleft, status->fs_blk_size);
    u32 mask;
    size_t adv1_len = oleft;

    assert (status->clu_nblocks * status->fs_blk_size
	    == status->clu_nchars);

    for (mask = 1; nBlksReqd != 0; mask <<= 1)
      {
	assert(mask != 0); /* test for overflow */
	if (obitmap & mask)
	  --nBlksReqd;
	else
	  adv1_len += status->fs_blk_size;
      }

    assert (tot_adv_len >= adv1_len);
    if (write_cluster (status, oleft, adv1_len, optr, obitmap, *offsetp))
      return -1;
    if (write_hole (zero_buf, tot_adv_len - adv1_len, status))
      return -1;
  }
  *offsetp += status->clu_nchars;

  return 0;
}


#if 0  /* Work in progress. */
static void 
process_dir (char *path, int len)
{
  DIR *dir = opendir (path);

  if (dir == NULL)
    return;  /* Fixme: should we check errno here, or give warning? */
  else 
    {
      struct dirent *de;

      path[len++] = '/';	/* fixme: we're assuming a path array. */
      path[len]   = '\0';

      while ((de = readdir (dir)) != NULL)
	if (de->d_name[0] != '.' 
	    || (de->d_name[1] != '\0' 
		&& (de->d_name[1] != '.' 
		    || de->d_name[2] != '\0')))
	  {
	    int newlen = len + strlen (de->d_name);

	    /* Make sure path[] has enough space. */
	    if (path_alloc <= newlen)
	      {
		/* fixme */
	      }
	        
	    strcpy (path + len, de->d_name);

	    /* fixme: lstat it.  

	       If regular file then e2[de]compress it.  
	       Elsif dir then [chattr {+|-}c and] recurse.
	       Else ignore it.

	       I don't know what the behaviour for directories should be: i.e. whether or not,
	       or in what circumstances, to chattr {+|-}c it. */
	  }

      closedir (dir);
    }
}
#endif


/* todo */
#define check_file(a,b,c,d) (fprintf(stderr, "check_file: nyi"), 1)


static int 
mf_process_file (file_status *status)
{
  unsigned long cluster;
  static unsigned char *ibuf = NULL;
  static unsigned char *obuf = NULL;

  gStatus = status;  /* Needed for signal handling. */
  /* todo: check for timing problems with signal handling. */

  if (streq (status->ipath, "-"))
    assert (0 && "stdio nyi");
  else
    if (mfo_open_files (status))
      return -1;

  if (heap_base == NULL && (heap_base = malloc (HEAP_SIZE)) == NULL)
    exit (fatal_error (status, "", "can't allocate heap", 0));

  /* Allocate working buffers. */
  {
    size_t bufsize;

    switch (status->action)
      {
      case ACT_compress:
	bufsize = status->clu_nchars;
	break;
      case ACT_decompress:
      case ACT_check:
	bufsize = EXT2_MAX_CLUSTER_BYTES;
	break;
      default:
	assert(0 && "durchfall");
      }

    if ((ibuf = realloc (ibuf, bufsize)) == NULL)
      exit (fatal_error (status, "", "can't allocate buffers", 0));

    if ((obuf = realloc (obuf, bufsize)) == NULL)
      exit (fatal_error (status, "", "can't allocate buffers", 0));
  }

  switch (status->action)
    {
    case ACT_compress:
      {
	off_t offset = 0;
	off_t ileft = status->ilen;

	for (cluster = 0;
	     cluster < round_up_div (status->ilen, status->clu_nchars);
	     cluster++)
	  if (mfl_process_cluster (cluster, status, ibuf, obuf, &ileft, &offset))
	    return -1;
      }
      break;
    case ACT_decompress:
    case ACT_check:
      {
	int rc;

	gMmap_len = (size_t) status->ilen;
	gMmap_start = mmap (NULL, gMmap_len, PROT_READ, MAP_PRIVATE, status->ifd, (off_t) 0);

	if (gMmap_start == (void *) (-1))
	  return file_error (status, status->ipath, ": can't mmap input file", errno);
	/* todo: Read if mmap fails.  (Modify decompress_stream()
	   first.) Change `munmap' below too. */

	if (status->action == ACT_decompress)
	  rc = decompress_stream (status, gMmap_start, ibuf, obuf);
	else
	  rc = check_file (status, gMmap_start, ibuf, obuf);

	gDefer_sig = true;
	do_munmap();
	gDefer_sig = false;
	if (gPending)
	  catch_signal (gPending);

	if (rc)
	  return rc;
      }
      break;
    }
  mfc_close_io (status);
  return 0; 
}


int 
main (int argc, char *argv[])
{
  int c, i;
  int rc = 0;	/* exit status */
  int n_files = 0;

  unsigned clu_nblocks = DFL_CLU_NBLOCKS;
  int compr_meth = DFL_COMPR_METH;
  enum action action;
  bool verbose = false;
  bool stress_test = false;
  bool do_continue = false;
  int fs_blk_size = 0;
  const char *suffix = "~z~";
  bool believe_bitmap = false;
  bool to_stdout = false;
  bool stdio_done = true;	/* todo: When stdio de/compression is
				   implemented, change this to
				   false. */
#ifdef BENCH
  bool do_bench = false;
  int nloop = 10;
#endif

  file_status *status_array = NULL;
  int n_status_alloc = 0;


  prog_name = xbasename (*argv);

  if ((strcmpi (prog_name, "e2uncompress") == 0)
      || (strcmpi (prog_name, "e2decompress") == 0))
    action = ACT_decompress;
  else
    action = ACT_compress;

  assert (sizeof (u32) == 4);
  assert (sizeof (u16) == 2);

  for (;;)
    {
      char *tmp;
      struct option longopts[] = {
	{"method", required_argument, NULL, (int) 'm'},
	{"algorithm", required_argument, NULL, (int) 'm'},
#ifdef BENCH
	{"benchmark", no_argument, NULL, (int) 'B'},
	{"loops", required_argument, NULL, (int) 'N'},
#endif
	{"force-blocksize", required_argument, NULL, (int) 'F'},
	{"suffix", required_argument, NULL, (int) 'S'},
	{"cluster-blocks", required_argument, NULL, (int) 'b'},
	{"check", no_argument, NULL, 0},
	{"decompress", no_argument, NULL, (int) 'd'},
	{"uncompress", no_argument, NULL, (int) 'd'},
	{"believe-bitmap", no_argument, NULL, 0},
	{"force", optional_argument, NULL, (int) 'f'},
	{"to-stdout", no_argument, NULL, (int) 'c'},
	{"stress-test", no_argument, NULL, (int) 0},
	{"help", no_argument, NULL, (int) 'h'},
	{"verbose", no_argument, NULL, (int) 'v'},
	{"version", no_argument, NULL, 0}
      };
      int option_ix = 0;

      c = getopt_long (argc, argv, "-m:A:BtF:N:S:b:cdfhv", longopts, &option_ix);
      if (c == EOF)
	{
	  assert (optind == argc);
	  break;
	}

      switch (c) 
	{
	case 0:
	  if (streq (longopts[option_ix].name, "version"))
	    return (fprintf (stderr, "%s\n", version_string) < 0);
	  if (streq (longopts[option_ix].name, "compress"))
	    {
	      action = ACT_compress;
	      break;
	    }
	  if (streq (longopts[option_ix].name, "believe-bitmap"))
	    {
	      believe_bitmap = true;
	      break;
	    }
	  if (streq (longopts[option_ix].name, "check"))
	    {
	      action = ACT_check;
	      break;
	    }
	  if (streq (longopts[option_ix].name, "stress-test"))
	    {
	      stress_test = true;
	      break;
	    }
	  (void) fprintf (stderr, "%s: bug: unrecognised option `%s'.  Please report.\n",
			  prog_name, longopts[option_ix].name);
	  abort ();

	case 'c': to_stdout = true; break;
	case 'd': action = ACT_decompress; break;
	case 'f': do_continue = 1; break;
	case 'v': verbose     = 1; break;
#ifdef BENCH
	case 'B': do_bench    = 1; break;
	case 'N':
	  nloop = strtol (optarg, &tmp, 0);
	  if (*tmp)
	    usage (1);
	  break;
#endif
	case 'F':
	  fs_blk_size = strtol (optarg, &tmp, 0);
	  if (*tmp)
	    usage (1);
	  if (fs_blk_size == 0)
	    break;  /* valid */
	  if ((!is_power_of_two(fs_blk_size))
	      || (fs_blk_size < EXT2_MIN_BLOCK_SIZE)
	      || (fs_blk_size > EXT2_MAX_BLOCK_SIZE))
	    {
	      fprintf (stderr,
		       "%s: invalid fs blocksize `%s'.\n"
		       "  Ought to be a power of two in the range ["
		       __STRING(EXT2_MIN_BLOCK_SIZE)
		       ", "
		       __STRING(EXT2_MAX_BLOCK_SIZE)
		       "].\n",
		       prog_name, argv[optind]);
	      exit (1);
	    }
	  break;

	case 'b':
	  clu_nblocks = strtol (optarg, &tmp, 0);
	  if (*tmp || (clu_nblocks !=  4 && clu_nblocks !=  8 &&
		       clu_nblocks != 16 && clu_nblocks != 32))
	    {
	      fprintf (stderr, "%s: invalid cluster size: %s\n",
		       prog_name, argv[optind]);
	      exit (1);
	    }
	  break;

	case 'S':
	  suffix = optarg;
	  break;

	case 'm':
	case 'A':
	  {
	    int x;

	    for (x = 0; name2id[x].name != NULL; x++)
	      if (strcasecmp (optarg, name2id[x].name) == 0)
		break;

	    if (name2id[x].name == NULL)
	      {
		fprintf (stderr, "%s: unknown algorithm: %s\n", prog_name, optarg);
		exit (1);
	      }

	    compr_meth = name2id[x].id;
	  }
	  break;

	case 'h':
	  usage (0);

	case 1:
	  /* Allocate more space for status_array if necessary. */
	  assert (n_files <= n_status_alloc);
	  if (n_files == n_status_alloc)
	    {
	      if (n_status_alloc)
		n_status_alloc *= 4;
	      else
		n_status_alloc = 1;

	      status_array = realloc (status_array, sizeof (file_status) * n_status_alloc);
	      if (status_array == NULL)
		{
		  (void) fprintf (stderr, "%s: failed to realloc %lu bytes for status array.\n",
				  prog_name, (unsigned long) n_status_alloc * sizeof (file_status));
		  return 1;
		}
	    }

	  if (streq (optarg, "-"))
	    if (stdio_done)
	      {
		(void) fprintf (stderr, "%s: Cannot process stdin"/*" twice"*/".\n", prog_name);
		return 1;
	      }
	    else
	      stdio_done = 1;
		
	  /* Copy current options to status. */
	  {
	    status_array[n_files].ipath = optarg;
	    status_array[n_files].opath = suffix;
	    status_array[n_files].fs_blk_size = fs_blk_size;
	    status_array[n_files].clu_nblocks = clu_nblocks;
	    status_array[n_files].compr_meth = compr_meth;
	    status_array[n_files].action = action;
	    status_array[n_files].believe_bitmap = believe_bitmap;
	    status_array[n_files].to_stdout = to_stdout;
	    status_array[n_files].stress_test = stress_test;
	    status_array[n_files].verbose = verbose;
	    status_array[n_files].do_continue = do_continue;
	    status_array[n_files].ifd = -1;
	    status_array[n_files].ofd = (to_stdout ? STDOUT_FILENO : -1);
#ifdef BENCH
	    if (do_bench)	
	      {
		if (action != ACT_compress)
		  {
		    (void) fprintf (stderr, 
				    "%s: benchmarking not supported with decompression.\n", 
				    prog_name);
		    (void) fprintf (stderr, "No files processed.\n");
		    return 1;
		  }
		status_array[n_files].do_bench = true;
		status_array[n_files].nloop = nloop;
	      }
	    else
	      {
		status_array[n_files].do_bench = false;
		status_array[n_files].nloop = 1;
	      }
#endif /* BENCH */
	  }

	  n_files++;
	  break;

	default:
	  usage (1);
	}
    }

  if (n_files == 0)
    usage (1);

  /* TODO: We could check if some options are unused. */

#ifdef BENCH
  if (do_bench)  /* fixme: should be: `If at least one file was processed with benchmarking info'. */
    {
      printf ("--------------------------------------\n");
# if 0
      /* tbi.  We used to print the options here, but that's a bit difficult if the options can vary
	 from file to file. */
      for (i = 0; i < optind; i++)
	printf ("%s ", argv[i]);
      printf ("\n\n");
# endif
      printf ("File name      Length            Ratio\n");
      printf ("---------      ------            -----\n");
    }
#endif

  /* Signal handling. */
  if ((signal (SIGTERM, catch_signal) == SIG_ERR)
      | (signal (SIGINT, catch_signal) == SIG_ERR)
      | (signal (SIGHUP, SIG_IGN) == SIG_ERR))
    (void) fprintf (stderr, "%s: Note: signal trapping failed.  (Continuing.)\n",
		    prog_name);

  /* Compress/decompress files. */
  for (i = 0; i < n_files; i++)
    if (mf_process_file (&status_array[i]))
      rc |= 1;

#ifdef BENCH
  if (do_bench) /* fixme: ditto, should only apply to benched files. */
    {
      printf ("---------      ------            -----\n");
      printf ("%-7s [%2d]  %7ld  %7ld   %5.1f",
	      algo[compr_meth].name, cluster_nblocks, ubyte, cbyte, 
	      ((ublk != 0) && (cblk != ublk)
	       ? (double) 100.0 * cblk / ublk
	       : 100.0));

      /* fixme: nloop in the below. */
      if (cmsec > 0)
	printf ("  %5.2f MB/s", (double) ubyte * nloop / (1000000.0 * cmsec));
      else
	printf ("  ??.?? MB/s");

      if (umsec > 0)
	printf (" %5.2f MB/s\n", (double) ubyte * nloop / (1000000.0 * umsec));
      else
	printf ("  ??.?? MB/s\n");
    }
#endif

  return 0;
}
